const ffmpeg = require('fluent-ffmpeg');
const fs = require('fs');
const path = require('path');
const sharp = require('sharp');

/**
 * 
 * @param {string} imagePath - Path to the image file
 * @param {number} duration - Duration in seconds to display the image
 * @param {string} outputClipPath - Path to the output clip file
 * @param {function} callback - Callback function
 */
function createClipFromImage(imagePath, duration, outputClipPath, callback) {
    ffmpeg(imagePath)
        .inputOptions('-loop 1')
        .inputOptions(`-t ${duration}`)
        .inputOptions('-framerate 1')
        .outputOptions([
            '-c:v libvpx',  // Use VP8 codec for video
            '-b:v 1M',  // Bitrate for video
            '-crf 10',  // Constant Rate Factor for quality
            '-auto-alt-ref 0',  // VP8 option to disable alternate reference frames
            '-c:a libvorbis'  // Use Vorbis codec for audio
        ])
        .output(outputClipPath)
        .on('end', () => {
            callback(null, outputClipPath); // Pass the clip path on success
        })
        .on('error', (err) => {
            console.error(`Error: ${err.message}`);
            callback(err);
        })
        .run();
}

function concatenateClips(clips, outputPath, callback) {
    const fileList = 'filelist.txt';
    const fileContent = clips.map(clip => `file '${clip}'`).join('\n');
    fs.writeFileSync(fileList, fileContent);
    console.log("created", outputPath, 'x' );
    ffmpeg()
        .input(fileList)
        .inputOptions(['-f concat', '-safe 0']) // Corrected input options
        .outputOptions('-c copy')
        .output(outputPath)
        .on('end', () => {
            console.log('Final video created successfully');
            callback(null);
            fs.unlinkSync(fileList);
            clips.forEach(clip => fs.unlinkSync(clip));
        })
        .on('error', (err) => {
            console.error(`Error: ${err}`);
            callback(err);
        })
        .run();
}

module.exports.createVideo = async function (screenshots, outputPath) {
    return new Promise(async (resolve, reject) => {
        const clips = [];

        try {
            for (let index = 0; index < screenshots.length; index++) {
                const screenshot = screenshots[index];
                const duration = screenshot.duration || 3;
                const clipPath = `clip${index}.webm`; // Temporary clip path

                const clip = await createClipFromImageAsync(screenshot, duration, clipPath);
                clips.push(clip);
            }

            await concatenateClipsAsync(clips, outputPath);
            resolve();
        } catch (error) {
            console.error('Error processing video:', error);
            reject(error);
        }
    });
};

// Helper function to convert callback-based createClipFromImage to promise-based
function createClipFromImageAsync(screenshot, duration, clipPath) {
    return new Promise((resolve, reject) => {
        createClipFromImage(screenshot, duration, clipPath, (err, clip) => {
            if (err) {
                reject(err);
            } else {
                resolve(clip);
            }
        });
    });
}

// Helper function to convert callback-based concatenateClips to promise-based
function concatenateClipsAsync(clips, outputPath) {
    return new Promise((resolve, reject) => {
        concatenateClips(clips, outputPath, (err) => {
            if (err) {
                reject(err);
            } else {
                resolve();
            }
        });
    });
}

module.exports.createVideoFromImages = async function (imageArr, options) {
    return new Promise(async (res, rej) => {
        let frameRate = options?.frameRate ? options.frameRate : 24;
        let tempImageDir = path.join(__dirname, '..', 'temporary', 'images');
        let outputFilePath = path.join(options.outputFilePath, `${options.outputFileName}.${options.outputFormat}`);
        let imagePathString = options?.imagePathString;
        let deleteImages = (options && options?.hasOwnProperty('deleteImages')) ? (!!options.deleteImages) : true;

        let frameNumber = 1;
        let frames = [];
        if (!imagePathString) { // if the path is not generated then generate the path
            imagePathString = path.join(tempImageDir, `${options.outputFileName}_%d.png`);
            ////// to create temporary image files //////
            for (let i = 0; i < imageArr.length; i++) {
                if (imageArr[i].filePath) {
                    let limit = imageArr[i]?.duration ? imageArr[i].duration : 1;
                    limit = limit * frameRate;
                    let frameNum = 0;
                    do {
                        let newFilePath = path.join(tempImageDir, `${options.outputFileName}_${frameNumber++}.png`);
                        await sharp(imageArr[i].filePath).toFile(newFilePath);
                        frames.push(newFilePath);
                    } while (++frameNum < limit);
                }
            }
        }
        ////// using ffmpeg creating video //////
        ffmpeg()
            .on('end', () => {
                res(path.join(options.outputFilePath, `${options.outputFileName}.${options.outputFormat}`));

                ////// to delete the frames that are created //////
                if (deleteImages)
                    frames.forEach((filePath) => {
                        fs.access(filePath, fs.constants.F_OK, (err) => {
                            if (!err) {
                                // File exists, proceed with deletion
                                fs.unlink(filePath, (err) => {
                                    if (err) { console.error(`Error deleting ${filePath}:`, err); }
                                });
                            };
                        });
                    });
            })
            .on('error', (err) => {
                console.error('Error creating video:', err);
                rej(undefined, err);
            })
            .input(imagePathString)
            .outputOptions('-c:v libvpx', '-b:v 1M', '-crf 10', '-auto-alt-ref 0', '-c:a libvorbis') // Added correct codecs
            .output(outputFilePath)
            .outputFPS(frameRate) // Set frames per second as needed
            .run();
    });
}

module.exports.joinAudioVideo = async function (videoPath, audioPath, output) {
    return new Promise((res, rej) => {
        ffmpeg()
            .input(videoPath)
            .input(audioPath)
            .output(output)
            .on('end', () => {
                res(output);
            })
            .on('error', (err) => {
                rej(err);
            })
            .run();
    });
}
